--Source for RGB/XYZ/Lab conversions and deltaE calculations: http://www.brucelindbloom.com/index.html?Equations.html

--## Choco ## --
-- 2016-23-01

local colors = {}
local colorPalette_Lab --(table) save lab colors when calc performed so only have to calc once 

--[[][][][][][][][][][][][][][][][][][][][][][]--
* Convert colors: RGB/Lab
* Calculate differences between two colors
* For non-palette color, finds closest palette color
* Simple matrix operations
--[][][][][][][][][][][][][][][][][][][][][][]]--

--Pre-calculated RGB2XYZ matrix for sRGB/D65
local sRGB = {
	{ 0.41245643908969, 0.35757607764391, 0.1804374832664 },
	{ 0.21267285140562, 0.71515215528782, 0.07217499330656 },
	{ 0.019333895582329, 0.1191920258813, 0.95030407853637 },
}

--Pre-calculated XYZ2RGB matrix for sRGB/D65
local sRGB_inv = {
	{ 3.2404541621141, -1.5371385127977, -0.49853140955602 },
	{ -0.96926603050519, 1.8760108454467, 0.04155601753035 },
	{ 0.055643430959115, -0.20402591351675, 1.0572251882232 },
}

--Reference illuminant tristimulus values CIE1931
local illuminants = {
	A = {1.09850, 1, 0.35585},
	B = {0.99072, 1, 0.85223},
	C = {0.98074, 1, 1.18232},
	D50 = {0.96422, 1, 0.82521},
	D55 = {0.95682, 1, 0.92149},
	D65 = {0.95047, 1, 1.08883}, --use this one
	D75 = {0.94972, 1, 1.22638},
	E = {1.00000, 1, 1.00000},
	F2 = {0.99186, 1, 0.67393},
	F7 = {0.95041, 1, 1.08747},
	F11 = {1.00962, 1, 0.64350},
}

local kappa = 24389/27
local epsilon = 216/24389

--// Returns determinant of a 2x2 matrix m (table)
function colors:determinant_2x2(m)
	assert(type(m)=="table" and #m == 2) --TODO error msg
	assert(type(m[1]=="table" and #m[1]==2 and type(m[2]=="table" and #m[2]==2))) --TODO error msg
	
	local m_1_1 = tonumber(m[1][1])
	local m_1_2 = tonumber(m[1][2])
	local m_2_1 = tonumber(m[2][1])
	local m_2_2 = tonumber(m[2][2])
	
	assert(m_1_1 and m_1_2 and m_2_1 and m_2_2) --TODO error msg
	
	return m_1_1*m_2_2 - m_1_2*m_2_1
end

--// Returns determinant of a 3x3 matrix m (table)
function colors:determinant_3x3(m)
	assert(type(m)=="table" and #m == 3) --TODO error msg
	assert(type(m[1]=="table" and #m[1]==3
		and m[2]=="table" and #m[2]==3
		and m[3]=="table" and #m[3]==3)) --TODO error msg
	
	local m_1_1 = tonumber(m[1][1])
	local m_1_2 = tonumber(m[1][2])
	local m_1_3 = tonumber(m[1][3])
	local m_2_1 = tonumber(m[2][1])
	local m_2_2 = tonumber(m[2][2])
	local m_2_3 = tonumber(m[2][3])
	local m_3_1 = tonumber(m[3][1])
	local m_3_2 = tonumber(m[3][2])
	local m_3_3 = tonumber(m[3][3])
	
	assert(m_1_1 and m_1_2 and m_1_3
		and m_2_1 and m_2_2 and m_2_3
		and m_3_1 and m_3_2 and m_3_3) --TODO error msg
	
	return m_1_1*(m_3_3*m_2_2-m_3_2*m_2_3) - m_2_1*(m_3_3*m_1_2-m_3_2*m_1_3) + m_3_1*(m_2_3*m_1_2-m_2_2*m_1_3)
end

--// Returns the inverse of the 2x2 matrix m (table)
function colors:matrix_inverse_2x2(m)
	local det = tonumber(self:determinant_2x2(m))
	assert(det) --TODO error msg
	
	if det~=0 then
		return {
			{m[2][2]/det, -1*m[1][2]/det},
			{-1*m[2][1]/det, m[1][1]/det},
		}
	else return false end
end

--// Returns the inverse of the 3x3 matrix m (table)
function colors:matrix_inverse_3x3(m)
	local det = tonumber(self:determinant_3x3(m))
	assert(det) --TODO error msg
	
	if det~=0 then
		local inv =  {
			{ m[3][3]*m[2][2]-m[3][2]*m[2][3], m[3][2]*m[1][3]-m[3][3]*m[1][2], m[2][3]*m[1][2]-m[2][2]*m[1][3] },
			{ m[3][1]*m[2][3]-m[3][3]*m[2][1], m[3][3]*m[1][1]-m[3][1]*m[1][3], m[2][1]*m[1][3]-m[2][3]*m[1][1] },
			{ m[3][2]*m[2][1]-m[3][1]*m[2][2], m[3][1]*m[1][2]-m[3][2]*m[1][1], m[2][2]*m[1][1]-m[2][1]*m[1][2] },
		}
		
		for _,row in ipairs(inv) do
			for n=1,3 do row[n]=row[n]/det end
		end
		
		return inv
	else return false end
end

--// used by XYZ2RGB operation
--// omit gamma value for sRGB
local function compand(val, gamma)
	val = tonumber(val)
	gamma = tonumber(gamma)
	assert(val) --TODO error msg
	assert(gamma~=0) --TODO error msg
	
	if not gamma then --sRGB only
		if val > 0.0031308 then
			val = 1.055*(val^(1/2.4)) - 0.055
		else val = 12.92*val end
	else val = val^(1/gamma) end
	
	return val
end

--// used by RGB2XYZ operation
--// omit gamma value for sRGB
local function invCompand(val, gamma)
	val = tonumber(val)
	gamma = tonumber(gamma)
	assert(val) --TODO error msg
	assert(gamma~=0) --TODO error msg
	
	if not gamma then --sRGB only
		if val > 0.04045 then
			val = ((val + 0.055)/1.055) ^ 2.4
		else val = val/12.92 end
	else val = val^gamma end
	
	return val
end

--// Convert RGB color to XYZ
--[[ Input r,g,b (red,green,blue values):
	*r,g,b are each integers from 0 to 255
	*if r is a string, omit g & b: e.g. "255 255 255"
	*if r is a table, omit g & b: e.g. {255, 255, 255}
	--TODO allow gvector input
]]
--[[ Returns table containing X,Y,Z values: e.g. {X, Y, Z}
	*where X,Y,Z are each numbers from 0 to 1
]]
function colors:rgb2xyz(r,g,b)
	if type(r)=="table" then
		r,g,b = r[1],r[2],r[3] --g & b inputs ignored
	elseif type(r)=="string" then r,g,b = r:match("^(%d+) (%d+) (%d+)") end
	
	r = tonumber(r)
	g = tonumber(g)
	b = tonumber(b)
	
	assert(r and g and b) --TODO error msg
	--TODO validate range of r,g,b values
	
	r = r/255
	g = g/255
	b = b/255
	
	r = invCompand(r)
	g = invCompand(g)
	b = invCompand(b)
	
	local xyz = {}
	for _,v in ipairs(sRGB) do
		table.insert(xyz, v[1]*r + v[2]*g + v[3]*b)
	end
	
	return xyz
end

--// Convert XYZ color values to CIE-Lab
--// input xyz is a table containing numbers in range 0-1: e.g. {x, y, z}
--// output is table {L, a, b}
function colors:xyz2lab(xyz)
	assert(type(xyz)=="table") --TODO error msg
	
	local f = {}
	for i=1,3 do
		local v = tonumber(xyz[i])
		assert(v) --TODO error msg
		
		local vref = v/illuminants.D65[i]
		
		if vref > epsilon then
			vref = vref ^ (1/3)
		else vref = (kappa*vref + 16)/116 end
		
		table.insert(f, vref)
	end
	
	local lab = {
		116*f[2] - 16,
		500*(f[1]-f[2]),
		200*(f[2]-f[3]),
	}
	
	return lab
end

--// Convert CIE-Lab color values to XYZ
--[[ Input L,a,b:
	*L,a,b are each numbers
	*if L is a table, omit a & b: e.g. {100, 0, 0}
]]
function colors:lab2xyz(L,a,b)
	if type(L)=="table" then L,a,b = L[1],L[2],L[3] end
	
	L = tonumber(L)
	a = tonumber(a)
	b = tonumber(b)
	
	assert(L and a and b) --TODO error msg
	
	local fy = (L+16)/116
	local f = {
		a/500 + fy,
		fy,
		fy - b/200,
	}
	
	local xyz = {}
	for i,v in ipairs(f) do
		local v3 = v^3
		if v3 <= epsilon then
			if i==2 then
				xyz[i] = L/kappa --reduction of redundant steps to reduce rounding error
			else xyz[i] = (116*v - 16)/kappa end
		else xyz[i] = v3 end
		
		xyz[i] = xyz[i]*illuminants.D65[i]
	end
	
	return xyz
end

--// Convert XYZ color values to RGB (returns TBD)
function colors:xyz2rgb(xyz)
	--TODO (not really needed)
end

function colors:rgb2lab(r,g,b) return self:xyz2lab(self:rgb2xyz(r,g,b)) end
function colors:lab2rgb(l,a,b) return self:xyz2rgb(self:lab2xyz(L,a,b)) end

function colors:calc_sRGB(white)
	local white = white or illuminants.D65
	
	local xr, yr = 0.64, 0.33
	local xg, yg = 0.30, 0.60
	local xb, yb = 0.15, 0.06
	
	local m = {
		{xr/yr, xg/yg, xb/yb},
		{1.0, 1.0, 1.0},
		{(1-xr-yr)/yr, (1-xg-yg)/yg, (1-xb-yb)/yb},
	}
	
	local m_inv = self:matrix_inverse_3x3(m)
	
	local s = {}
	for i,row in ipairs(m_inv) do
		local sum = 0
		for n=1,3 do sum = sum + (row[n]*white[n]) end
		s[i] = sum
	end
	
	local xyz = {}
	for i,row in ipairs(m) do
		xyz[i] = {}
		for n,v in ipairs(row) do
			xyz[i][n] = s[n]*v
		end
	end
	
	return xyz
	--return self:matrix_inverse_3x3(xyz)
end

--// Returns the perceptual color difference between two Lab colors as a positive number (smaller numbers are a closer match)
--// deltaE1994 and deltaE2000 have better accuracy but are more computationally intense
function colors:deltaE1976(lab1, lab2)
	if type(lab1)=='userdata' and string.match(tostring(lab1), '^gvector ') then lab1 = {lab1:x(), lab1:y(), lab1:z()} end
	if type(lab2)=='userdata' and string.match(tostring(lab2), '^gvector ') then lab2 = {lab2:x(), lab2:y(), lab2:z()} end
	assert(type(lab1)=="table") --TODO error msg
	assert(type(lab2)=="table") --TODO error msg
	
	--TODO assert table values as numbers
	
	local dL = lab1[1] - lab2[1]
	local da = lab1[2] - lab2[2]
	local db = lab1[3] - lab2[3]
	
	return math.sqrt(dL*dL + da*da + db*db)
end

--// Returns the perceptual color difference between two Lab colors as a positive number (smaller numbers are a closer match)
--// deltaE2000 has better accuracy but is more computationally intense
function colors:deltaE1994(lab1, lab2)
	if type(lab1)=='userdata' and string.match(tostring(lab1), '^gvector ') then lab1 = {lab1:x(), lab1:y(), lab1:z()} end
	if type(lab2)=='userdata' and string.match(tostring(lab2), '^gvector ') then lab2 = {lab2:x(), lab2:y(), lab2:z()} end
	assert(type(lab1)=="table") --TODO error msg
	assert(type(lab2)=="table") --TODO error msg
	
	local dL = lab1[1] - lab2[1]
	local da = lab1[2] - lab2[2]
	local db = lab1[3] - lab2[3]
	
	local c1 = math.sqrt(lab1[2]*lab1[2] + lab1[3]*lab1[3])
	local c2 = math.sqrt(lab2[2]*lab2[2] + lab2[3]*lab2[3])
	local dC = c1 - c2
	local dH2 = da*da + db*db - dC*dC
	
	local kC = 1 + 0.045*c1
	local kH = 1 + 0.015*c1
	
	return math.sqrt(dL*dL + dC*dC/kC/kC + dH2/kH/kH)
end

--// Returns the perceptual color difference between two Lab colors as a positive number (smaller numbers are a closer match)
function colors:deltaE2000(lab1, lab2)
	if type(lab1)=='userdata' and string.match(tostring(lab1), '^gvector ') then lab1 = {lab1:x(), lab1:y(), lab1:z()} end
	if type(lab2)=='userdata' and string.match(tostring(lab2), '^gvector ') then lab2 = {lab2:x(), lab2:y(), lab2:z()} end
	assert(type(lab1)=="table") --TODO error msg
	assert(type(lab2)=="table") --TODO error msg
	
	local lBarPrime = 0.5*(lab1[1] + lab2[1])
	
	local c1 = math.sqrt(lab1[2]*lab1[2] + lab1[3]*lab1[3])
	local c2 = math.sqrt(lab2[2]*lab2[2] + lab2[3]*lab2[3])
	local cBar = 0.5*(c1+c2)
	local cBar7 = cBar*cBar*cBar*cBar*cBar*cBar*cBar
	
	local g = 0.5*(1 - math.sqrt(cBar7/(cBar7 + 6103515625))) --6103515625 = 25^7
	local a1Prime = lab1[2]*(1+g)
	local a2Prime = lab2[2]*(1+g)
	
	local c1Prime = math.sqrt(a1Prime*a1Prime + lab1[3]*lab1[3])
	local c2Prime = math.sqrt(a2Prime*a2Prime + lab2[3]*lab2[3])
	
	local cBarPrime = 0.5*(c1Prime + c2Prime)
	local cBarPrime7 = cBarPrime * cBarPrime * cBarPrime * cBarPrime * cBarPrime * cBarPrime * cBarPrime
	
	local h1Prime = 180*math.atan2(lab1[3], a1Prime)/math.pi
	local h2Prime = 180*math.atan2(lab2[3], a2Prime)/math.pi
	if h1Prime < 0 then h1Prime = h1Prime + 360 end
	if h2Prime < 0 then h2Prime = h2Prime + 360 end
	local hBarPrime = 0.5*(h1Prime + h2Prime) + (math.abs(h1Prime - h2Prime)>180 and 180 or 0)
	local dTheta = 30*math.exp(-1 * ((hBarPrime-275)/25) * ((hBarPrime-275)/25))
	
	local t = 1 -
		0.17 * math.cos(math.pi * ((hBarPrime - 30)/180) ) +
		0.24 * math.cos(math.pi * (hBarPrime/90) ) +
		0.32 * math.cos(math.pi * ((hBarPrime + 2)/60) ) -
		0.20 * math.cos(math.pi * ((hBarPrime - 15.75)/45) )
	
	local dhPrime = h2Prime - h1Prime + ( math.abs(h2Prime - h1Prime)<=180 and 0 or (h2Prime<=h1Prime and 360 or -360) )
	local dHPrime = 2*math.sqrt(c1Prime*c2Prime) * math.sin(math.pi*(dhPrime/360))
	local dLPrime = lab2[1] - lab1[1]
	local dCPrime = c2Prime - c1Prime
	
	local sL = 1 + 0.015*(lBarPrime - 50)*(lBarPrime - 50)/math.sqrt(20 + (lBarPrime - 50)*(lBarPrime - 50))
	local sC = 1 + 0.045*cBarPrime
	local sH = 1 + 0.015*cBarPrime * t
	
	local rC = math.sqrt(cBarPrime7 / (cBarPrime7 + 6103515625)) --6103515625 = 25^7
	local rT = -2*rC * math.sin(math.pi*(dTheta/90))
	
	return math.sqrt(
		(dLPrime/sL) * (dLPrime/sL) + 
		(dCPrime/sC) * (dCPrime/sC) +
		(dHPrime/sH) * (dHPrime/sH) +
		(dCPrime/sC) * (dHPrime/sH) * rT)
end

--// Calculates Lab color values for every palette color and saves to colorPalette_Lab (if not already calculated)
function colors:calcLabPalette()
	if not colorPalette_Lab then
		colorPalette_Lab = {}
		for i,rgb in ipairs(ShipPalette_string) do
			colorPalette_Lab[i] = self:rgb2lab(rgb)
		end
		return true
	else return end
end

--// Finds palette color that is closest perceptual match to given rgb color
--// Input is an RGB color as a string (e.g. "255 255 255"), table (e.g. {255, 255, 255}) or as 3 values
--// Returns palette index (integer from 1 to 256) and deltaE value for how close the match is (smaller numbers are a closer match)
function colors:nearestPaletteColor(r,g,b)
	local lab = self:rgb2lab(r,g,b) --convert to lab color
	if not colorPalette_Lab then self:calcLabPalette() end --convert all palette colors to lab if not done already
	
	local best_i --index of closest matching palette color
	local best_dE --dE value corresponding to closest match
	
	--Compare to each palette color
	for i,lab2 in ipairs(colorPalette_Lab) do
		local dE = self:deltaE2000(lab, lab2)
		
		if not best_dE or dE<best_dE then
			best_i = i
			best_dE = dE
		end			
	end
	
	return best_i, best_dE
end

--// Calculates deltaE value for 2 colors
--// Input is an RGB color as a string (e.g. "255 255 255"), Lab colors as a table (e.g. {L, a, b}) or as a palette color (1-256) if a number
function colors:calc_deltaE(color1, color2)
	if not colorPalette_Lab then self:calcLabPalette() end --convert all palette colors to lab if not done already
	
	local deltaE = colors.deltaE2000 --which deltaE function to use
	
	local palette_color1 = tonumber(color1)
	local palette_color2 = tonumber(color2)
	
	local lab1
	local lab2
	
	if palette_color1 then
		lab1 = colorPalette_Lab[palette_color1]
	elseif type(color1)=="string" then --RGB color
		lab1 = self:rgb2lab(color1)
	elseif type(color1)=="table" then --Lab color
		lab1 = color1
	end
	
	if palette_color2 then
		lab2 = colorPalette_Lab[palette_color2]
	elseif type(color2)=="string" then --RGB color
		lab2 = self:rgb2lab(color2)
	elseif type(color2)=="table" then --Lab color
		lab2 = color2
	end
	
	return deltaE(self, lab1, lab2)
end

--TEST list matching palette colors for each palette color
--[[local tol = 15
for i,rgb in ipairs(ShipPalette_string) do
	local matches = {}
	for j,rgb2 in ipairs(ShipPalette_string) do
		local dE = colors:calc_deltaE(i,j)
		if dE<=tol then table.insert(matches, j) end
	end
	console_print(i..": ",table.concat(matches, ", "))
end]]

return colors

